class Config:
  INFINITY = "float('inf')"
  FACTORIAL = "math.factorial"

  FLOOR = "math.floor"
  CEIL = "math.ceil"

  SQRT = "math.sqrt"

  SIN = "math.sin"
  COS = "math.cos"
  TAN = "math.tan"
  SINH = "math.sinh"
  COSH = "math.cosh"
  TANH = "math.tanh"

  ASIN = "math.asin"
  ACOS = "math.acos"
  ATAN = "math.atan"
  ASINH = "math.asinh"
  ACOSH = "math.acosh"
  ATANH = "math.atanh"

  LOG = "math.log"
  EXP = "math.exp"
  ABS = "abs"

  E = "math.e"
